# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ===========================================================================

from __future__ import absolute_import as __absolute_import

from minorminer.minorminer import miner, VARORDER, find_embedding

from minorminer.package_info import __version__, __author__, __authoremail__, __description__



# Copyright 2019 - 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import absolute_import as __absolute_import
from minorminer._minorminer import miner, VARORDER, find_embedding as __find_embedding
from functools import wraps as __wraps

# This wrapper exists to overcome a curious limitation of Cython, and make
# find_embedding friendlier for the inspect module.
@__wraps(__find_embedding)
def find_embedding(S, T,
                   max_no_improvement=10,
                   random_seed=None,
                   timeout=1000,
                   max_beta=None,
                   tries=10,
                   inner_rounds=None,
                   chainlength_patience=10,
                   max_fill=None,
                   threads=1,
                   return_overlap=False,
                   skip_initialization=False,
                   verbose=0,
                   interactive=False,
                   initial_chains=(),
                   fixed_chains=(),
                   restrict_chains=(),
                   suspend_chains=(),
                   ):
    return __find_embedding(S, T,
                            max_no_improvement=max_no_improvement,
                            random_seed=random_seed,
                            timeout=timeout,
                            max_beta=max_beta,
                            tries=tries,
                            inner_rounds=inner_rounds,
                            chainlength_patience=chainlength_patience,
                            max_fill=max_fill,
                            threads=threads,
                            return_overlap=return_overlap,
                            skip_initialization=skip_initialization,
                            verbose=verbose,
                            interactive=interactive,
                            initial_chains=initial_chains,
                            fixed_chains=fixed_chains,
                            restrict_chains=restrict_chains,
                            suspend_chains=suspend_chains,
                            )



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from dwave_networkx.generators.zephyr import zephyr_graph
import networkx as nx

from minorminer.busclique import busgraph_cache

__all__ = ['find_clique_embedding', 'find_biclique_embedding']


def _get_target_graph(m=None, target_graph=None):
    if target_graph is None:
        if m is None:
            raise TypeError("m and target_graph cannot both be None.")
        target_graph = zephyr_graph(m)
    return target_graph


@nx.utils.decorators.nodes_or_number(0)
def find_clique_embedding(k, m=None, target_graph=None):
    """Find an embedding for a clique in a Zephyr graph.

    Given a clique (fully connected graph) and target Zephyr graph, attempts
    to find an embedding.

    Args:
        k (int/iterable/:obj:`networkx.Graph`): A complete graph to embed,
            formatted as a number of nodes, node labels, or a NetworkX graph.

        m (int): Number of tiles in a row of a square Zephyr graph.  Required
            to generate an ``m``-by-``m`` Zephyr graph when ``target_graph`` is None.

        target_graph (:obj:`networkx.Graph`): A Zephyr graph.  Required when
            ``m`` is None.

    Returns:
        dict: An embedding as a dict, where keys represent the clique's nodes
        and values, formatted as lists, represent chains of zephyr coordinates.

    Examples:
        This example finds an embedding for a :math:`K_5` complete graph in a
        2-by-2 Zephyr graph.

        >>> from dwave.embedding.zephyr import find_clique_embedding
        >>> find_clique_embedding(5, 2)
        {0: (16, 96), 1: (18, 98), 2: (20, 100), 3: (22, 102), 4: (24, 104)}
    """
    _, nodes = k
    g = _get_target_graph(m, target_graph)
    embedding = busgraph_cache(g).find_clique_embedding(nodes)

    if len(embedding) != len(nodes):
        raise ValueError("No clique embedding found")

    return embedding


@nx.utils.decorators.nodes_or_number(0)
@nx.utils.decorators.nodes_or_number(1)
def find_biclique_embedding(a, b, m=None, target_graph=None):
    """Find an embedding for a biclique in a Zephyr graph.

    Given a biclique (a bipartite graph where every vertex in a set in
    connected to all vertices in the other set) and a target :term:`Zephyr`
    graph, attempts to find an embedding.

    Args:
        a (int/iterable):
            Describes the left shore of the biclique to embed.  If ``a`` is an
            integer, the left shore will be labelled [0, a-1].  If ``a`` is an
            iterable, the left shore will be labelled by ``a``.

        b (int/iterable):
            Describes the right shore of the biclique to embed.  If ``b`` is an
            integer and ``a`` is an iterable, the right shore will be labelled
            [0, b-1].  If both ``a`` and ``b`` are integers, the right shore
            will be labelled [a, a+b-1].  If ``b`` is an iterable, the right
            shore will be labelled by ``b``.

        m (int): Number of tiles in a row of a square Zephyr graph.  Required to
            generate an ``m``-by-``m`` Zephyr graph when ``target_graph`` is None.

        target_graph (:obj:`networkx.Graph`): A Zephyr graph.  Required when ``m``
            is None.

    Returns:
        tuple: A 2-tuple containing:

            dict: An embedding mapping the left shore of the biclique to
            the Zephyr lattice.

            dict: An embedding mapping the right shore of the biclique to
            the Zephyr lattice.

    Examples:
        This example finds an embedding for an alphanumerically labeled
        biclique in a 2x2 Zephyr graph.

        >>> from dwave.embedding.zephyr import find_biclique_embedding
        >>> left, right = find_biclique_embedding(['a', 'b', 'c'], ['d', 'e'], 2)
        >>> print(left, right)
        {'a': (0,), 'b': (4,), 'c': (8,)} {'d': (80,), 'e': (84,)}
    """
    _a, anodes = a
    _b, bnodes = b

    if isinstance(_a, int) and isinstance(_b, int):
        bnodes = [len(anodes) + x for x in bnodes]

    if set(anodes).intersection(set(bnodes)):
        raise ValueError("a and b overlap")

    g = _get_target_graph(m, target_graph)
    embedding = busgraph_cache(g).find_biclique_embedding(len(anodes), len(bnodes))

    if not embedding:
        raise ValueError("No biclique embedding found")

    return ({x: embedding[anodes.index(x)] for x in anodes},
            {y: embedding[bnodes.index(y) + len(anodes)] for y in bnodes})



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

import networkx as nx

from minorminer.utils import MissingChainError, ChainOverlapError, \
    DisconnectedChainError, InvalidNodeError, MissingEdgeError

__all__ = ['diagnose_embedding', 'is_valid_embedding', 'verify_embedding']


def diagnose_embedding(emb, source, target):
    """Diagnose a minor embedding.

    Produces a generator that lists all issues with the embedding. User-friendly
    variants of this function are :func:`is_valid_embedding`, which returns a
    bool, and :func:`verify_embedding`, which raises the first observed error.

    Args:
        emb (dict):
            A mapping of source nodes to arrays of target nodes as a dict
            of form {s: [t, ...], ...}, where s is a source-graph variable and t
            is a target-graph variable.

        source (list/:obj:`networkx.Graph`):
            Graph to be embedded as a NetworkX graph or a list of edges.

        target (list/:obj:`networkx.Graph`):
            Graph being embedded into as a NetworkX graph or a list of edges.

    Yields:
        Errors yielded in the form `ExceptionClass, arg1, arg2,...`, where the
        arguments following the class are used to construct the exception object,
        which are subclasses of :exc:`.EmbeddingError`.

            :exc:`.MissingChainError`, snode: a source node label that does not
            occur as a key of `emb`, or for which emb[snode] is empty.

            :exc:`.ChainOverlapError`, tnode, snode0, snode1: a target node which
            occurs in both `emb[snode0]` and `emb[snode1]`.

            :exc:`.DisconnectedChainError`, snode: a source node label whose chain
            is not a connected subgraph of `target`.

            :exc:`.InvalidNodeError`, tnode, snode: a source node label and putative
            target node label that is not a node of `target`.

            :exc:`.MissingEdgeError`, snode0, snode1: a pair of source node labels
            defining an edge that is not present between their chains.

    Examples:
        This example diagnoses an invalid embedding from a triangular source graph
        to a square target graph. A valid embedding, such as
        `emb = {0: [1], 1: [0], 2: [2, 3]}`, yields no errors.

         >>> from dwave.embedding import diagnose_embedding
         >>> import networkx as nx
         >>> source = nx.complete_graph(3)
         >>> target = nx.cycle_graph(4)
         >>> embedding = {0: [2], 1: [1, 'a'], 2: [2, 3]}
         >>> diagnosis = diagnose_embedding(embedding, source, target)
         >>> for problem in diagnosis:  # doctest: +SKIP
         ...     print(problem)
         (<class 'dwave.embedding.exceptions.InvalidNodeError'>, 1, 'a')
         (<class 'dwave.embedding.exceptions.ChainOverlapError'>, 2, 2, 0)

    """

    if not hasattr(source, 'edges'):
        source = nx.Graph(source)
    if not hasattr(target, 'edges'):
        target = nx.Graph(target)

    labels = {}
    embedded = set()
    overlaps = set()
    for x in source:
        try:
            embx = emb[x]
            missing_chain = len(embx) == 0
        except KeyError:
            missing_chain = True
        if missing_chain:
            yield MissingChainError, x
            continue

        all_present = True
        for q in embx:
            if q not in target:
                all_present = False
                yield InvalidNodeError, x, q
            elif x not in labels.setdefault(q, {x}):
                labels[q].add(x)
                overlaps.add(q)

        if all_present:
            embedded.add(x)
            if not nx.is_connected(target.subgraph(embx)):
                yield DisconnectedChainError, x

    for q in overlaps:
        nodes = list(labels[q])
        root = nodes[0]
        for x in nodes[1:]:
            yield ChainOverlapError, q, root, x

    yielded = nx.Graph()
    for p, q in target.subgraph(labels).edges():
        yielded.add_edges_from((x, y) for x in labels[p] for y in labels[q])

    for x, y in source.edges():
        if x == y:
            continue
        if x in embedded and y in embedded and not yielded.has_edge(x, y):
            yield MissingEdgeError, x, y


def is_valid_embedding(emb, source, target):
    """A simple (bool) diagnostic for minor embeddings.

    See :func:`diagnose_embedding` for a more detailed diagnostic and more information.

    Args:
        emb (dict): A mapping of source nodes to arrays of target nodes as a dict
            of form {s: [t, ...], ...}, where s is a source-graph variable and t
            is a target-graph variable.
        source (graph or edgelist): Graph to be embedded.
        target (graph or edgelist): Graph being embedded into.

    Returns:
        bool: True if `emb` is valid.

    """
    for _ in diagnose_embedding(emb, source, target):
        return False
    return True


def verify_embedding(emb, source, target, ignore_errors=()):
    """A simple (exception-raising) diagnostic for minor embeddings.

    See :func:`diagnose_embedding` for a more detailed diagnostic and more information.

    Args:
        emb (dict): A mapping of source nodes to arrays of target nodes as a dict
            of form {s: [t, ...], ...}, where s is a source-graph variable and t
            is a target-graph variable.
        source (graph or edgelist): Graph to be embedded
        target (graph or edgelist): Graph being embedded into

    Raises:
        EmbeddingError: A catch-all class for the following errors:

            MissingChainError: A key is missing from `emb` or the associated chain is empty.

            ChainOverlapError: Two chains contain the same target node.

            DisconnectedChainError: A chain is disconnected.

            InvalidNodeError: A chain contains a node label not found in `target`.

            MissingEdgeError: A source edge is not represented by any target edges.

    Returns:
        bool: True if no exception is raised.
    """

    for error in diagnose_embedding(emb, source, target):
        eclass = error[0]
        if eclass not in ignore_errors:
            raise eclass(*error[1:])
    return True
